# This file is a part of the Hypothesis-Cats Python package.
#
# Copyright (C) 2022  Paul Wolneykien <manowar@altlinux.org>
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301, USA.

"""
Hypothesis-Cats allows you to classify values generated by a
`Hypothesis <https://hypothesis.readthedocs.io/en/latest/>`_
strategy. This module defines utility classes for category
descriptors.
"""

from typing import Optional, Mapping, Any, Union, Sequence

class Cat():
    """
    A simple base category descriptor consisting of a category
    name and a comment.
    """

    def __init__(self, name: str = None,
                 comment: Optional[str] = None,
                 tags: Optional[Union[str, Sequence[str]]] = None):
        """
        :param name: The name of the category.

        :param comment: The comment or a description for it.

        :param tags: The set of tags assigned to this category.

        :raises ValueError: If name is empty.
        """
        if not name:
            raise ValueError('Category name should not be empty!')
        self.name = name
        self.comment = comment

        if tags:
            if isinstance(tags, str):
                self.tags = [ tags ]
            else:
                self.tags = [ *tags ]
        else:
            self.tags = []

    def __repr__(self) -> str:
        """
        Returns the string representation of a category: its
        name is angle brackets.

        :return: Name of the category in angle brackets.
        """
        return '<%s>' % self.name

    def __str__(self) -> str:
        """
        The default string representation of a category ommits
        the comment.

        :return: Name of the category.
        """
        return self.name

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]) -> 'Cat':
        """
        Tries to create the :class:`Cat` from a dictionary.
        The dictionary has to define a value under the ``'name'`` key.
        The optional ``'comment'`` and ``'tags'`` values, if such
        keys are present, are also used. All other values in the
        dictionary are silently ignored.

        :param d: A dictionary with at least ``'name'`` and,
            optionally, ``'comment'`` and ``'tags'`` keys.

        :return: The fresh :class:`Cat:` instance with name and
            comment taken from the dictionary.

        :raises KeyError: If there is no ``'name'`` key in the
            supplied dictionary.

        :raises ValueError: If the value ``'name'`` is empty.
        """
        _d = {}
        _d['name'] = d['name']
        if 'comment' in d:
            _d['comment'] = d['comment']
        if 'tags' in d:
            _d['tags'] = d['tags']
        return cls(**_d)
